﻿#LoiLoNote School Sideloading Setup Version 2.0
param(
    [ValidateSet('Select', 'Install','Uninstall','ShowLog','ProvisionedInstall','ProvisionedUninstall', 'ElevatedProvisionedInstall', 'ElevatedProvisionedUninstall', 'RemoteInstall')]
    [string] $Action = 'Select',
    [string] $Packages,
    [switch] $Always,
    [switch] $Force,
    [switch] $EnsureInstall
)

$ErrorActionPreference = "Stop"

Set-PSDebug -Strict

$DefaultPackageName = "LoiLoinc.51956F723A3DD.s1"
$CompanyName = "LoiLo"
$ScriptPath = (Get-Variable MyInvocation).Value.MyCommand.Path
$ScriptDirPath = Split-Path $ScriptPath -Parent
$ScriptName = $(Get-ChildItem $ScriptPath).BaseName
$WorkDirPath = Join-Path "$env:LOCALAPPDATA" (Join-Path $CompanyName $ScriptName)
$TempDirPath = Join-Path $WorkDirPath "Temp"
$LogDirPath = Join-Path $WorkDirPath "Logs"

#region Log

$LogMaxLines = 1000
$LogFileName = "log.txt"
$LogFilePath = Join-Path $LogDirPath $LogFileName
$LogBackupFileName = "log.txt.back"
$LogBackupFilePath = Join-Path $LogDirPath $LogBackupFileName

function Append-Log{

    param(
        [ValidateSet("INFO", "ERROR", ignorecase=$true)]
        [string] $type,
        [string] $msg
    )

    $now = Get-Date -Format "yyyy/MM/dd HH:mm:ss"
    $message = ("$now {0,-5} $msg" -f $type)

    if ($type -ne "ERROR") {
        Write-Host $message
    } else {
        Write-Host $message -ForegroundColor Red
    }

    if(-not (Test-Path $LogDirPath -PathType Container)){
        New-Item $LogDirPath -ItemType Directory -Force | Out-Null
    }

    if(Test-Path $LogFilePath -PathType Leaf){
        $line = (Get-Content $LogFilePath | Measure-Object –Line).Lines
        if( $line -ge $LogMaxLines ){
            if(Test-Path $LogBackupFilePath -PathType Leaf){
                Remove-Item $LogBackupFilePath -Force 
            }        
            Rename-Item $LogFilePath -NewName $LogBackupFileName 
        }
    }

    Write-Output $message | Out-File ${LogFilePath} -Append -Encoding utf8
}

#endregion

#region Zip

function Expand-Zip {
    param([string] $zipFilePath, [string] $dstDirPath)

    New-Item $dstDirPath -ItemType Directory -Force | Out-NUll

    # Zipファイルにファイルおよびフォルダを投入
    $shell = New-Object -com Shell.Application
    $zip = $shell.NameSpace($zipFilePath)
    foreach($item in $zip.items())
    {
        $shell.Namespace($dstDirPath).copyhere($item)
    }
}

#endregion 

#region Logic

function Find-File {
    param( [string] $dirPath, [string] $filePattern)

    $queryPath = (Join-Path $dirPath "*.*")
    $foundItems = @(Get-ChildItem $queryPath -Include $filePattern | Where-Object { $_.Mode -NotMatch "d" })
    if($foundItems.Length -le 0) {
        return $null
    }
    return $foundItems[0]
}

function Get-TargetPackageInfo {
    param([string] $dirName)
    $packageJsonFilePath = Join-Path $dirName "package.json"
    $packageJsonContent = Get-Content -Path $packageJsonFilePath -Encoding UTF8 -Raw 
    $info = ConvertFrom-Json $packageJsonContent
    return $info
}

function Compare-Version {
    param([string] $version1, [string] $version2)

    $array1 = @($version1 -split "." , 0 , "simplematch")
    $array2 = @($version2 -split "." , 0 , "simplematch")

    for($i = 0; $i -lt 4; $i++) {
        $contains1 = ($array1.Count - 1) -ge $i;
        $contains2 = ($array2.Count - 1) -ge $i;
        if((-not $contains1) -and (-not $contains2)) {
            return 0;
        } elseif(-not $contains1) {
            return -1;
        } elseif(-not $contains2) {
            return 1;
        }

        $num1 = [int]$array1[$i]
        $num2 = [int]$array2[$i]

        if($num1 -gt $num2) {
            return 1;
        } elseif($num1 -lt $num2) {
            return -1;
        }
    }

    return 0;
}

function Get-DependencyPackages {
    param([string] $dirPath)

    $dependenciesDirPath = (Join-Path $dirPath "Dependencies")
    $dependencyPackages = @()
    if (Test-Path $dependenciesDirPath) {
        $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "*.appx") | Where-Object { $_.Mode -NotMatch "d" }

        if (($Env:Processor_Architecture -eq "x86" -or $Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x86"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x86\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        } 
        if (($Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x64"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x64\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
        if (($Env:Processor_Architecture -eq "arm") -and (Test-Path (Join-Path $dependenciesDirPath "arm"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "arm\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
    }

    return $dependencyPackages
}

function Install-Package {
    param([string] $dirPath)
    
    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    if ($dependencyPackages.FullName.Count -gt 0) {
        Append-Log INFO "Found dependency($dependencyPackages)"
        Add-AppxPackage "$bundleFilePath" -DependencyPath $dependencyPackages.FullName -ForceApplicationShutdown
    } else {
        Add-AppxPackage "$bundleFilePath" -ForceApplicationShutdown
    }
}

function Install-ProvisionedPackage {
    param([string] $dirPath)
    
    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    if ($dependencyPackages.FullName.Count -gt 0) {
        Append-Log INFO "Found dependency($dependencyPackages)"
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -DependencyPackagePath $dependencyPackages.FullName -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    } else {
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    }
}

function Uninstall-Package {
    param([string] $packageName)
    $package = Get-AppxPackage -Name $packageName
    if($package) {
        Append-Log INFO "Uninstalling..."
        Remove-AppxPackage -Package $package.PackageFullName
        Append-Log INFO "Complete!"
    } else {
        Append-Log INFO "Not installed."
    }
}

function Uninstall-ProvisionedPackage{
    param([string] $packageName)
    $packages = @(Get-AppxProvisionedPackage -Online | Where DisplayName -eq $packageName)
    if($packages) {
        foreach( $package in $packages ) {
            $packageFullname = $package.PackageName
            Append-Log INFO "Uninstalling ProvisionedPackage($packageFullname)..."
            $image = Remove-AppxProvisionedPackage -Online -PackageName $packageFullname
            Append-Log INFO ($image | Format-List | Out-String)
        }
        Append-Log INFO "Complete!" 
    } else {
        Append-Log INFO "Not installed."
    }   
}

function Run-RemoteInstall {

    if($Env:Processor_Architecture -eq "arm") {
        throw "ARM is not supported RemoteInstall."
    }

    $remoteTempDirPath = Join-Path $TempDirPath "Remote"
    New-Item $remoteTempDirPath -ItemType Directory -Force | Out-NUll
    $remotePackageInfoFilePath = Join-Path $remoteTempDirPath "package.json"

    Append-Log INFO "Checking remote version..."

    $remotePackageInfoUri = "http://assets.loilo.tv/loilonote/windows/sideloading/package.json"
    Invoke-RestMethod $remotePackageInfoUri -OutFile $remotePackageInfoFilePath
    $packageJsonContent = Get-Content -Path $remotePackageInfoFilePath -Encoding UTF8 -Raw 
    $targetPackageInfo = ConvertFrom-Json $packageJsonContent
    $packageName = $targetPackageInfo.Name
    $targetVersion = $targetPackageInfo.Version
    $targetPublisherId = $targetPackageInfo.PublisherId

    Append-Log INFO "Remote version=$targetVersion"

    $installedPackage = Get-AppxPackage -Name $packageName

    $removeRequired = $false
    if($installedPackage) {
        $compareResult = Compare-Version $installedPackage.Version $targetVersion
        if($compareResult -lt 0) {
            Append-Log INFO ("Currently installed $packageName version=" + $installedPackage.Version + "(Older).")

            $installedPackagePublisherId = $installedPackage.PublisherId
            if($installedPackagePublisherId -ne $targetPublisherId) {
                Append-Log INFO "PublisherId changed.($installedPackagePublisherId to $targetPublisherId)"
                $removeRequired = $true
            }
        } elseif($compareResult -gt 0) {
            Append-Log INFO ("Currently installed $packageName version=" + $installedPackage.Version + "(Newer).")
            if(-not $Always) {
                return
            }
            $removeRequired = $true
        } else {
            Append-Log INFO ("Already installed $packageName version=" + $installedPackage.Version + ".")
            return
        }               
    } else {
        Append-Log INFO "Currently not installed $packageName."
    }

    Append-Log INFO "Downloading version=$targetVersion ..."

    $remotePackageFileName = "LoiLoNoteSetup_$targetVersion.zip"
    $remotePackageFilePath = Join-Path $remoteTempDirPath $remotePackageFileName

    $remotePackageUri = "http://assets.loilo.tv/loilonote/windows/sideloading/$remotePackageFileName"
    Invoke-RestMethod $remotePackageUri -OutFile $remotePackageFilePath

    $remoteUnzipDirPath = Join-Path $remoteTempDirPath "LoiLoNoteSetup"
    New-Item $remoteUnzipDirPath -ItemType Directory -Force | Out-NUll
    Expand-Zip $remotePackageFilePath $remoteUnzipDirPath

    $packageDirPath = Join-Path $remoteUnzipDirPath "Packages"

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Not exists $indexFilePath ."
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
        throw "Not exists $targetDirPath ."
    }

    if( $removeRequired ) {
        Append-Log INFO "Removing installed version..."
        Uninstall-Package $packageName
    }

    Append-Log INFO "Installing version=$targetVersion ..."
    Install-Package $targetDirPath
    Append-Log INFO "Complete!"
}

function Run-Install {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Not exists $indexFilePath ."
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
        throw "Not exists $targetDirPath ."
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    $targetVersion = $targetPackageInfo.Version
    $targetPublisherId = $targetPackageInfo.PublisherId

    $installedPackage = Get-AppxPackage -Name $packageName

    $removeRequired = $false
    if($installedPackage) {
        $compareResult = Compare-Version $installedPackage.Version $targetVersion
        if($compareResult -lt 0) {
            Append-Log INFO ("Currently installed $packageName version=" + $installedPackage.Version + "(Older).")

            $installedPackagePublisherId = $installedPackage.PublisherId
            if($installedPackagePublisherId -ne $targetPublisherId) {
                Append-Log INFO "PublisherId changed.($installedPackagePublisherId to $targetPublisherId)"
                $removeRequired = $true
            }
        } elseif($compareResult -gt 0) {
            Append-Log INFO ("Currently installed $packageName version=" + $installedPackage.Version + "(Newer).")
            if(-not $Always) {
                return
            }
            $removeRequired = $true
        } else {
            Append-Log INFO ("Already installed $packageName version=" + $installedPackage.Version + ".")
            return
        }               
    } else {
        Append-Log INFO "Currently not installed $packageName."
    }

    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    if( $removeRequired ) {
        Append-Log INFO "Removing installed version..."
        Uninstall-Package $packageName
    }

    Append-Log INFO "Installing version=$targetVersion ..."
    Install-Package $copyDirPath
    Append-Log INFO "Complete!"
}

function Run-ProvisionedInstall {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Not exists $indexFilePath ."
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
        throw "Not exists $targetDirPath ."
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    Uninstall-ProvisionedPackage $packageName

    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    Append-Log INFO "Installing ProvisionedPackage..."
    Install-ProvisionedPackage $copyDirPath
    Append-Log INFO "Complete!"
}

function Ensure-TempDir {
    Append-Log INFO "Ready $TempDirPath"
    if(Test-Path $TempDirPath)
    {
        Remove-Item $TempDirPath -Recurse -Force
    }
    New-Item $TempDirPath -ItemType Directory -Force | Out-NUll
}

function Run{

    $running = $Action;
    if($running -eq 'Select') {
        $input = [int](Read-Host "Please select option below.`n0:Exit, 1:Install, 2:Uninstall, 3:ShowLog, 4:ProvisionedInstall, 5:ProvisionedUninstall, 6:RemoteInstall`n=>")
        if($input -eq 1) {
            $running = 'Install';
        } elseif ($input -eq 2) {
            $running = 'Uninstall';
        } elseif ($input -eq 3) {
            $running = 'ShowLog';
        } elseif ($input -eq 4) {
            $running = 'ProvisionedInstall';
        } elseif ($input -eq 5) {
            $running = 'ProvisionedUninstall';
        } elseif ($input -eq 6) {
            $running = 'RemoteInstall';
        } else {
            Append-Log INFO "Exit."
            return $true
        }
    }

    Append-Log INFO "Action=$running"

    if($running -eq 'Install') {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
        Run-Install $packageDirPath

        return $false
    } elseif ($running -eq 'Uninstall') {
        Uninstall-Package $DefaultPackageName

        return $false
    } elseif ($running -eq 'ShowLog') {
        Start-Process "$LogDirPath" | Out-Null

        return $true
    } elseif ($running -eq 'ProvisionedInstall') {

        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedInstall'

        if($Force) {
            $relaunchArgs += ' -Force'
        }

        if( $Packages -ne "" ) {
            $relaunchArgs += ' -Packages "' + $Packages + '"'
        }
        
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq 'ProvisionedUninstall') {
        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedUninstall'
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq "ElevatedProvisionedInstall") {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
        Run-ProvisionedInstall $packageDirPath

        return $false
    } elseif ($running -eq "ElevatedProvisionedUninstall"){
        Uninstall-ProvisionedPackage $DefaultPackageName

        return $false
    } elseif ($running -eq "RemoteInstall"){
        Ensure-TempDir
        Append-Log INFO "Param( Always=$Always )"

        try {            
            Run-RemoteInstall
        } catch {
            Append-Log ERROR $_.Exception.Message

            if($EnsureInstall){
                
                Append-Log INFO "Installing local package..."

                Ensure-TempDir
                $packageDirPath = Join-Path $ScriptDirPath "Packages"
                if( $Packages -ne "" ){
                    $packageDirPath = $Packages
                }

                $Always = $false
                Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
                Run-Install $packageDirPath
            }
        }

        return $false
    } 

    return $false
}

#endregion


#
# Main script entry point
#

Append-Log INFO "[Start] $ScriptName"

$SilentExit = $false
try {
    $SilentExit = Run
} catch {
    Append-Log ERROR $_.Exception.Message
} finally {
    try {
        if(Test-Path $TempDirPath)
        {
            Append-Log INFO "CleanUp $TempDirPath"
            Remove-Item $TempDirPath -Recurse -Force
        }
    } catch {
        Append-Log ERROR $_.Exception.Message
    }

    Append-Log INFO "[End] $ScriptName"
    if ((-not $SilentExit) -and (-not $Force)) {
        Pause
    }
}

# SIG # Begin signature block
# MIIVDAYJKoZIhvcNAQcCoIIU/TCCFPkCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQU1BjTA5yqQ2ujDAko5+CrjLmJ
# vDKggg/8MIIEmTCCA4GgAwIBAgIPFojwOSVeY45pFDkH5jMLMA0GCSqGSIb3DQEB
# BQUAMIGVMQswCQYDVQQGEwJVUzELMAkGA1UECBMCVVQxFzAVBgNVBAcTDlNhbHQg
# TGFrZSBDaXR5MR4wHAYDVQQKExVUaGUgVVNFUlRSVVNUIE5ldHdvcmsxITAfBgNV
# BAsTGGh0dHA6Ly93d3cudXNlcnRydXN0LmNvbTEdMBsGA1UEAxMUVVROLVVTRVJG
# aXJzdC1PYmplY3QwHhcNMTUxMjMxMDAwMDAwWhcNMTkwNzA5MTg0MDM2WjCBhDEL
# MAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UE
# BxMHU2FsZm9yZDEaMBgGA1UEChMRQ09NT0RPIENBIExpbWl0ZWQxKjAoBgNVBAMT
# IUNPTU9ETyBTSEEtMSBUaW1lIFN0YW1waW5nIFNpZ25lcjCCASIwDQYJKoZIhvcN
# AQEBBQADggEPADCCAQoCggEBAOnpPd/XNwjJHjiyUlNCbSLxscQGBGue/YJ0UEN9
# xqC7H075AnEmse9D2IOMSPznD5d6muuc3qajDjscRBh1jnilF2n+SRik4rtcTv6O
# KlR6UPDV9syR55l51955lNeWM/4Og74iv2MWLKPdKBuvPavql9LxvwQQ5z1IRf0f
# aGXBf1mZacAiMQxibqdcZQEhsGPEIhgn7ub80gA9Ry6ouIZWXQTcExclbhzfRA8V
# zbfbpVd2Qm8AaIKZ0uPB3vCLlFdM7AiQIiHOIiuYDELmQpOUmJPv/QbZP7xbm1Q8
# ILHuatZHesWrgOkwmt7xpD9VTQoJNIp1KdJprZcPUL/4ygkCAwEAAaOB9DCB8TAf
# BgNVHSMEGDAWgBTa7WR0FJwUPKvdmam9WyhNizzJ2DAdBgNVHQ4EFgQUjmstM2v0
# M6eTsxOapeAK9xI1aogwDgYDVR0PAQH/BAQDAgbAMAwGA1UdEwEB/wQCMAAwFgYD
# VR0lAQH/BAwwCgYIKwYBBQUHAwgwQgYDVR0fBDswOTA3oDWgM4YxaHR0cDovL2Ny
# bC51c2VydHJ1c3QuY29tL1VUTi1VU0VSRmlyc3QtT2JqZWN0LmNybDA1BggrBgEF
# BQcBAQQpMCcwJQYIKwYBBQUHMAGGGWh0dHA6Ly9vY3NwLnVzZXJ0cnVzdC5jb20w
# DQYJKoZIhvcNAQEFBQADggEBALozJEBAjHzbWJ+zYJiy9cAx/usfblD2CuDk5oGt
# Joei3/2z2vRz8wD7KRuJGxU+22tSkyvErDmB1zxnV5o5NuAoCJrjOU+biQl/e8Vh
# f1mJMiUKaq4aPvCiJ6i2w7iH9xYESEE9XNjsn00gMQTZZaHtzWkHUxY93TYCCojr
# QOUGMAu4Fkvc77xVCf/GPhIudrPczkLv+XZX4bcKBUCYWJpdcRaTcYxlgepv84n3
# +3OttOe/2Y5vqgtPJfO44dXddZhogfiqwNGAwsTEOYnB9smebNd0+dmX+E/CmgrN
# Xo/4GengpZ/E8JIh5i15Jcki+cPwOoRXrToW9GOUEB1d0MYwggV3MIIEX6ADAgEC
# AhBWpZFYm+Qv4q5lkeb9Une7MA0GCSqGSIb3DQEBCwUAMH0xCzAJBgNVBAYTAkdC
# MRswGQYDVQQIExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNVBAcTB1NhbGZvcmQx
# GjAYBgNVBAoTEUNPTU9ETyBDQSBMaW1pdGVkMSMwIQYDVQQDExpDT01PRE8gUlNB
# IENvZGUgU2lnbmluZyBDQTAeFw0xNTAxMjMwMDAwMDBaFw0xNzAxMjIyMzU5NTla
# MIHFMQswCQYDVQQGEwJKUDERMA8GA1UEEQwIMjMxLTAwMDYxHjAcBgNVBAgMFU91
# dHNpZGUgVW5pdGVkIFN0YXRlczEUMBIGA1UEBwwLWW9rb2hhbWFzaGkxDDAKBgNV
# BAkMAzQwMzEjMCEGA1UECQwaTmFrYWt1IE1pbmFtaW5ha2Fkb3JpIDQgNDMxEjAQ
# BgNVBAoMCUxvaUxvIGluYzESMBAGA1UECwwJTG9pTG8gaW5jMRIwEAYDVQQDDAlM
# b2lMbyBpbmMwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDnfQjPPsVq
# DE2+yob+ctESmTVgBC8dTeVHSHn3alTxcx0sr7Dgp1YBAfp1CnmwKXRYSk3ZKhor
# UMLTXBburPuPa3VTjxu2SOnppD0htfuvA1CGisOCPXN5D1dQnJwq0AMzZ049nsAy
# sIlystXbVD4sBxs7TugE+YdjhhsEX4yvA21q/uqbJe9yOaAQ2qzH+5m70zIaKc3t
# SAGms2KgaHp3zqYqoBDODMdNfpQX3xsTKIrNZrxyJ/w5a63WeltMYE3EYdRXFSmB
# O5RCj9w8bIWyhoTPIFTSfNgO2OxYK6dWedgSKjbvyC9h52oSUSIp3Fr44FS8xc7a
# L9yPs1nehWHtAgMBAAGjggGoMIIBpDAfBgNVHSMEGDAWgBQpkWD/ik366/mmarjP
# +eZLvUnOEjAdBgNVHQ4EFgQUXGiZv0bAEqJEXf8i9887SDBXZdMwDgYDVR0PAQH/
# BAQDAgeAMAwGA1UdEwEB/wQCMAAwEwYDVR0lBAwwCgYIKwYBBQUHAwMwEQYJYIZI
# AYb4QgEBBAQDAgQQMEYGA1UdIAQ/MD0wOwYMKwYBBAGyMQECAQMCMCswKQYIKwYB
# BQUHAgEWHWh0dHBzOi8vc2VjdXJlLmNvbW9kby5uZXQvQ1BTMEMGA1UdHwQ8MDow
# OKA2oDSGMmh0dHA6Ly9jcmwuY29tb2RvY2EuY29tL0NPTU9ET1JTQUNvZGVTaWdu
# aW5nQ0EuY3JsMHQGCCsGAQUFBwEBBGgwZjA+BggrBgEFBQcwAoYyaHR0cDovL2Ny
# dC5jb21vZG9jYS5jb20vQ09NT0RPUlNBQ29kZVNpZ25pbmdDQS5jcnQwJAYIKwYB
# BQUHMAGGGGh0dHA6Ly9vY3NwLmNvbW9kb2NhLmNvbTAZBgNVHREEEjAQgQ5sb2ls
# b0Bsb2lsby50djANBgkqhkiG9w0BAQsFAAOCAQEAh87sJw6Rz9WkR4xH7GO7Ogg6
# 0qIM4unPGmvnu6LUHok3rfdu/VL/gSsCi5AkQcZoTQsiFSYUrffz8JYOm/Lm0bT8
# SF2glQ5r8tPodIgfo/X/7Svw7pmTO8kxpEbh/oZ5dVsIHHuXdDlS140VdaZj6Fge
# ZsKGgrECG8Y8AcxqQElewUOyRHW25xRyWj1R3yTz/Thv8urU+sE/4ELEPLJcmzUp
# EexzpfI7z+ctiKJbu21zXLcV9R18wUSrsJV1KASb/KUNI95HlIeoECg7lzvqEMmg
# 8C9xvxd+0q31lFd126dPv710ZlbirXAIuLtUKhrwc5uVHB5YufvrCwsJCOjtyTCC
# BeAwggPIoAMCAQICEC58h8wOk0pS/pT9HLfNNK8wDQYJKoZIhvcNAQEMBQAwgYUx
# CzAJBgNVBAYTAkdCMRswGQYDVQQIExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNV
# BAcTB1NhbGZvcmQxGjAYBgNVBAoTEUNPTU9ETyBDQSBMaW1pdGVkMSswKQYDVQQD
# EyJDT01PRE8gUlNBIENlcnRpZmljYXRpb24gQXV0aG9yaXR5MB4XDTEzMDUwOTAw
# MDAwMFoXDTI4MDUwODIzNTk1OVowfTELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdy
# ZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9yZDEaMBgGA1UEChMRQ09N
# T0RPIENBIExpbWl0ZWQxIzAhBgNVBAMTGkNPTU9ETyBSU0EgQ29kZSBTaWduaW5n
# IENBMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAppiQY3eRNH+K0d3p
# ZzER68we/TEds7liVz+TvFvjnx4kMhEna7xRkafPnp4ls1+BqBgPHR4gMA77YXuG
# CbPj/aJonRwsnb9y4+R1oOU1I47Jiu4aDGTH2EKhe7VSA0s6sI4jS0tj4CKUN3vV
# eZAKFBhRLOb+wRLwHD9hYQqMotz2wzCqzSgYdUjBeVoIzbuMVYz31HaQOjNGUHOY
# XPSFSmsPgN1e1r39qS/AJfX5eNeNXxDCRFU8kDwxRstwrgepCuOvwQFvkBoj4l84
# 28YIXUezg0HwLgA3FLkSqnmSUs2HD3vYYimkfjC9G7WMcrRI8uPoIfleTGJ5iwIG
# n3/VCwIDAQABo4IBUTCCAU0wHwYDVR0jBBgwFoAUu69+Aj36pvE8hI6t7jiY7Nky
# MtQwHQYDVR0OBBYEFCmRYP+KTfrr+aZquM/55ku9Sc4SMA4GA1UdDwEB/wQEAwIB
# hjASBgNVHRMBAf8ECDAGAQH/AgEAMBMGA1UdJQQMMAoGCCsGAQUFBwMDMBEGA1Ud
# IAQKMAgwBgYEVR0gADBMBgNVHR8ERTBDMEGgP6A9hjtodHRwOi8vY3JsLmNvbW9k
# b2NhLmNvbS9DT01PRE9SU0FDZXJ0aWZpY2F0aW9uQXV0aG9yaXR5LmNybDBxBggr
# BgEFBQcBAQRlMGMwOwYIKwYBBQUHMAKGL2h0dHA6Ly9jcnQuY29tb2RvY2EuY29t
# L0NPTU9ET1JTQUFkZFRydXN0Q0EuY3J0MCQGCCsGAQUFBzABhhhodHRwOi8vb2Nz
# cC5jb21vZG9jYS5jb20wDQYJKoZIhvcNAQEMBQADggIBAAI/AjnD7vjKO4neDG1N
# sfFOkk+vwjgsBMzFYxGrCWOvq6LXAj/MbxnDPdYaCJT/JdipiKcrEBrgm7EHIhpR
# HDrU4ekJv+YkdK8eexYxbiPvVFEtUgLidQgFTPG3UeFRAMaH9mzuEER2V2rx31hr
# IapJ1Hw3Tr3/tnVUQBg2V2cRzU8C5P7z2vx1F9vst/dlCSNJH0NXg+p+IHdhyE3y
# u2VNqPeFRQevemknZZApQIvfezpROYyoH3B5rW1CIKLPDGwDjEzNcweU51qOOgS6
# oqF8H8tjOhWn1BUbp1JHMqn0v2RH0aofU04yMHPCb7d4gp1c/0a7ayIdiAv4G6o0
# pvyM9d1/ZYyMMVcx0DbsR6HPy4uo7xwYWMUGd8pLm1GvTAhKeo/io1Lijo7MJuSy
# 2OU4wqjtxoGcNWupWGFKCpe0S0K2VZ2+medwbVn4bSoMfxlgXwyaiGwwrFIJkBYb
# /yud29AgyonqKH4yjhnfe0gzHtdl+K7J+IMUk3Z9ZNCOzr41ff9yMU2fnr0ebC+o
# jwwGUPuMJ7N2yfTm18M04oyHIYZh/r9VdOEhdwMKaGy75Mmp5s9ZJet87EUOeWZo
# 6CLNuO+YhU2WETwJitB/vCgoE/tqylSNklzNwmWYBp7OSFvUtTeTRkF8B93P+kPv
# umdh/31J4LswfVyA4+YWOUunMYIEejCCBHYCAQEwgZEwfTELMAkGA1UEBhMCR0Ix
# GzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9yZDEa
# MBgGA1UEChMRQ09NT0RPIENBIExpbWl0ZWQxIzAhBgNVBAMTGkNPTU9ETyBSU0Eg
# Q29kZSBTaWduaW5nIENBAhBWpZFYm+Qv4q5lkeb9Une7MAkGBSsOAwIaBQCgeDAY
# BgorBgEEAYI3AgEMMQowCKACgAChAoAAMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3
# AgEEMBwGCisGAQQBgjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEW
# BBTTn+Bvi0/8SsKLuCXj+i/ADc2F5zANBgkqhkiG9w0BAQEFAASCAQCJFNNkfJ4k
# DZVy+XVkvEsHGwUEX4o4HABxNBaJjiH0nlHoBZcWlX/a3DvM90LK9nTkMis+xdbl
# UAWbxalhSbQUhC7EPRF/H0486591UdNAngZi7fcGMZxJs/TCWdh+QxHTLBif6eHY
# xHBeSVzSmW3RTcPA/VCCQKTpHwUTbTTv7txZEcsaf5WYHKmvyoL4QcqR0TD3rOml
# g3MpeagDGQQ4yAP9CMKhhBTNk3XHtDiwzXz/45RpLW5t8f28Ins/nJJvK970pKM2
# e1pcW4G3Tw/aPrRhCUbPTZDMh/EOiSyOZRY3/176tduN92PBDWTHcFkKLITQAMKr
# FqwpH0MveCSjoYICQzCCAj8GCSqGSIb3DQEJBjGCAjAwggIsAgEAMIGpMIGVMQsw
# CQYDVQQGEwJVUzELMAkGA1UECBMCVVQxFzAVBgNVBAcTDlNhbHQgTGFrZSBDaXR5
# MR4wHAYDVQQKExVUaGUgVVNFUlRSVVNUIE5ldHdvcmsxITAfBgNVBAsTGGh0dHA6
# Ly93d3cudXNlcnRydXN0LmNvbTEdMBsGA1UEAxMUVVROLVVTRVJGaXJzdC1PYmpl
# Y3QCDxaI8DklXmOOaRQ5B+YzCzAJBgUrDgMCGgUAoF0wGAYJKoZIhvcNAQkDMQsG
# CSqGSIb3DQEHATAcBgkqhkiG9w0BCQUxDxcNMTYxMDEyMDcxMDA0WjAjBgkqhkiG
# 9w0BCQQxFgQUS+9DUQaWYYTeecXPDpDBqNZ1LMIwDQYJKoZIhvcNAQEBBQAEggEA
# ZqwmqU7mRLeHe1fPqrc992DP5fNei2ZO84e0Ew3w1avrzbgTD5yhCsxG2+Jfymfg
# LQcmcflgAdqjrSiwTvejucaoU4OMDGEWhU1hhShKxvp1VPJYTD7CZt5uIZmTaM1Z
# gSc946o92PJlZvnjjssuK7Mcb8CUXKppiAiHJsQbDubk0OogdQ+dl0RteB5Vo1m5
# TFzq1HpFjh8zvCcwceMjDQzXMEokXnOGCGMnO/6oUkMY8lIkHd6pOpP3vHKD83PP
# t7O3yATVe+4F2bjwdizAeS6uT1UpYVHC9UgtUqGM3kDtuFZshXBMgtRQ/2ajvhEU
# 4B6ZeNt8Wbvf8QrPcSDyoA==
# SIG # End signature block
